/*
 * AFM 88593 encoding.
 *
 * This file is automatically generated from file `88593.txt'.  If you
 * have any corrections to this file, please, edit file `88593.txt' instead.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "afmint.h"

AFMEncodingTable afm_88593_encoding[] =
{
  {0x00, AFM_ENC_NONE},
  {0x01, AFM_ENC_NONE},
  {0x02, AFM_ENC_NONE},
  {0x03, AFM_ENC_NONE},
  {0x04, AFM_ENC_NONE},
  {0x05, AFM_ENC_NONE},
  {0x06, AFM_ENC_NONE},
  {0x07, AFM_ENC_NONE},
  {0x08, AFM_ENC_NONE},
  {0x09, AFM_ENC_NONE},
  {0x0a, AFM_ENC_NONE},
  {0x0b, AFM_ENC_NONE},
  {0x0c, AFM_ENC_NONE},
  {0x0d, AFM_ENC_NONE},
  {0x0e, AFM_ENC_NONE},
  {0x0f, AFM_ENC_NONE},
  {0x10, AFM_ENC_NONE},
  {0x11, AFM_ENC_NONE},
  {0x12, AFM_ENC_NONE},
  {0x13, AFM_ENC_NONE},
  {0x14, AFM_ENC_NONE},
  {0x15, AFM_ENC_NONE},
  {0x16, AFM_ENC_NONE},
  {0x17, AFM_ENC_NONE},
  {0x18, AFM_ENC_NONE},
  {0x19, AFM_ENC_NONE},
  {0x1a, AFM_ENC_NONE},
  {0x1b, AFM_ENC_NONE},
  {0x1c, AFM_ENC_NONE},
  {0x1d, AFM_ENC_NONE},
  {0x1e, AFM_ENC_NONE},
  {0x1f, AFM_ENC_NONE},
  {0x20, "space"},
  {0x21, "exclam"},
  {0x22, "quotedbl"},
  {0x23, "numbersign"},
  {0x24, "dollar"},
  {0x25, "percent"},
  {0x26, "ampersand"},
  {0x27, "quoteright"},
  {0x28, "parenleft"},
  {0x29, "parenright"},
  {0x2a, "asterisk"},
  {0x2b, "plus"},
  {0x2c, "comma"},
  {0x2d, "hyphen"},
  {0x2e, "period"},
  {0x2f, "slash"},
  {0x30, "zero"},
  {0x31, "one"},
  {0x32, "two"},
  {0x33, "three"},
  {0x34, "four"},
  {0x35, "five"},
  {0x36, "six"},
  {0x37, "seven"},
  {0x38, "eight"},
  {0x39, "nine"},
  {0x3a, "colon"},
  {0x3b, "semicolon"},
  {0x3c, "less"},
  {0x3d, "equal"},
  {0x3e, "greater"},
  {0x3f, "question"},
  {0x40, "at"},
  {0x41, "A"},
  {0x42, "B"},
  {0x43, "C"},
  {0x44, "D"},
  {0x45, "E"},
  {0x46, "F"},
  {0x47, "G"},
  {0x48, "H"},
  {0x49, "I"},
  {0x4a, "J"},
  {0x4b, "K"},
  {0x4c, "L"},
  {0x4d, "M"},
  {0x4e, "N"},
  {0x4f, "O"},
  {0x50, "P"},
  {0x51, "Q"},
  {0x52, "R"},
  {0x53, "S"},
  {0x54, "T"},
  {0x55, "U"},
  {0x56, "V"},
  {0x57, "W"},
  {0x58, "X"},
  {0x59, "Y"},
  {0x5a, "Z"},
  {0x5b, "bracketleft"},
  {0x5c, "backslash"},
  {0x5d, "bracketright"},
  {0x5e, "asciicircum"},
  {0x5f, "underscore"},
  {0x60, "quoteleft"},
  {0x61, "a"},
  {0x62, "b"},
  {0x63, "c"},
  {0x64, "d"},
  {0x65, "e"},
  {0x66, "f"},
  {0x67, "g"},
  {0x68, "h"},
  {0x69, "i"},
  {0x6a, "j"},
  {0x6b, "k"},
  {0x6c, "l"},
  {0x6d, "m"},
  {0x6e, "n"},
  {0x6f, "o"},
  {0x70, "p"},
  {0x71, "q"},
  {0x72, "r"},
  {0x73, "s"},
  {0x74, "t"},
  {0x75, "u"},
  {0x76, "v"},
  {0x77, "w"},
  {0x78, "x"},
  {0x79, "y"},
  {0x7a, "z"},
  {0x7b, "braceleft"},
  {0x7c, "bar"},
  {0x7d, "braceright"},
  {0x7e, "tilde"},
  {0x7f, AFM_ENC_NONE},
  {0x80, AFM_ENC_NONE},
  {0x81, AFM_ENC_NONE},
  {0x82, AFM_ENC_NONE},
  {0x83, AFM_ENC_NONE},
  {0x84, AFM_ENC_NONE},
  {0x85, AFM_ENC_NONE},
  {0x86, AFM_ENC_NONE},
  {0x87, AFM_ENC_NONE},
  {0x88, AFM_ENC_NONE},
  {0x89, AFM_ENC_NONE},
  {0x8a, AFM_ENC_NONE},
  {0x8b, AFM_ENC_NONE},
  {0x8c, AFM_ENC_NONE},
  {0x8d, AFM_ENC_NONE},
  {0x8e, AFM_ENC_NONE},
  {0x8f, AFM_ENC_NONE},
  {0x90, AFM_ENC_NONE},
  {0x91, AFM_ENC_NONE},
  {0x92, AFM_ENC_NONE},
  {0x93, AFM_ENC_NONE},
  {0x94, AFM_ENC_NONE},
  {0x95, AFM_ENC_NONE},
  {0x96, AFM_ENC_NONE},
  {0x97, AFM_ENC_NONE},
  {0x98, AFM_ENC_NONE},
  {0x99, AFM_ENC_NONE},
  {0x9a, AFM_ENC_NONE},
  {0x9b, AFM_ENC_NONE},
  {0x9c, AFM_ENC_NONE},
  {0x9d, AFM_ENC_NONE},
  {0x9e, AFM_ENC_NONE},
  {0x9f, AFM_ENC_NONE},
  {0xa0, "space"},
  {0xa1, "Hstroke"},
  {0xa2, "breve"},
  {0xa3, "sterling"},
  {0xa4, "currency"},
  {0xa5, "yen"},
  {0xa6, "Hcircumflex"},
  {0xa7, "section"},
  {0xa8, "dieresis"},
  {0xa9, "Idot"},
  {0xaa, "Scedilla"},
  {0xab, "Gbreve"},
  {0xac, "Jcircumflex"},
  {0xad, "hyphen"},
  {0xae, "registered"},
  {0xaf, "Zdot"},
  {0xb0, "degree"},
  {0xb1, "hstroke"},
  {0xb2, "twosuperior"},
  {0xb3, "threesuperior"},
  {0xb4, "acute"},
  {0xb5, "mu"},
  {0xb6, "hcircumflex"},
  {0xb7, "bullet"},
  {0xb8, "cedilla"},
  {0xb9, "dotlessi"},
  {0xba, "scedilla"},
  {0xbb, "gbreve"},
  {0xbc, "jcircumflex"},
  {0xbd, "onehalf"},
  {0xbe, "threequarters"},
  {0xbf, "zdot"},
  {0xc0, "Agrave"},
  {0xc1, "Aacute"},
  {0xc2, "Acircumflex"},
  {0xc3, "Atilde"},
  {0xc4, "Adieresis"},
  {0xc5, "Cdot"},
  {0xc6, "Ccircumflex"},
  {0xc7, "Ccedilla"},
  {0xc8, "Egrave"},
  {0xc9, "Eacute"},
  {0xca, "Ecircumflex"},
  {0xcb, "Edieresis"},
  {0xcc, "Igrave"},
  {0xcd, "Iacute"},
  {0xce, "Icircumflex"},
  {0xcf, "Idieresis"},
  {0xd0, "Eth"},
  {0xd1, "Ntilde"},
  {0xd2, "Ograve"},
  {0xd3, "Oacute"},
  {0xd4, "Ocircumflex"},
  {0xd5, "Gdot"},
  {0xd6, "Odieresis"},
  {0xd7, "multiply"},
  {0xd8, "Gcircumflex"},
  {0xd9, "Ugrave"},
  {0xda, "Uacute"},
  {0xdb, "Ucircumflex"},
  {0xdc, "Udieresis"},
  {0xdd, "Ubreve"},
  {0xde, "Scircumflex"},
  {0xdf, "germandbls"},
  {0xe0, "agrave"},
  {0xe1, "aacute"},
  {0xe2, "acircumflex"},
  {0xe3, "atilde"},
  {0xe4, "adieresis"},
  {0xe5, "cdot"},
  {0xe6, "ccircumflex"},
  {0xe7, "ccedilla"},
  {0xe8, "egrave"},
  {0xe9, "eacute"},
  {0xea, "ecircumflex"},
  {0xeb, "edieresis"},
  {0xec, "igrave"},
  {0xed, "iacute"},
  {0xee, "icircumflex"},
  {0xef, "idieresis"},
  {0xf0, "eth"},
  {0xf1, "ntilde"},
  {0xf2, "ograve"},
  {0xf3, "oacute"},
  {0xf4, "ocircumflex"},
  {0xf5, "gdot"},
  {0xf6, "odieresis"},
  {0xf7, "divide"},
  {0xf8, "gcircumflex"},
  {0xf9, "ugrave"},
  {0xfa, "uacute"},
  {0xfb, "ucircumflex"},
  {0xfc, "udieresis"},
  {0xfd, "ubreve"},
  {0xfe, "scircumflex"},
  {0xff, "dotaccent"},
  {-1,   NULL},
};
